import datetime
import json
import math
import os
import os.path as osp

import pandas


def parse(
    root_dir,
    sort=None,
    q=None,
    float_format=None,
    params_basename="params.json",
    log_basename="log.csv",
    key_epoch="epoch",
    key_iteration="iteration",
    key_elapsed_time="elapsed_time",
):
    """parse(root_dir, sort=None, q=None)

    Parse log directory to create a log table.

    Parameters
    ----------
    root_dir: str
        Root directory that contains logs.
    sort: str, optional
        Sort logs.
    q: str, optional
        Query of log keys.
    float_format: str, optional.
        Float value formatting (default: '{:.3g}').

    Returns
    -------
    df: pandas.DataFrame
        Log table.
    summary_keys: list of str
        List of keys from summary.
    args_keys: list of str
        List of keys from args.
    log_keys: list of str
        List of keys from log.
    """
    if float_format is None:
        float_format = "{:.3g}"

    log_dirs = sorted(os.listdir(root_dir))

    # params
    params_keys = set()
    log_keys = set()
    data = []
    for log_dir in sorted(log_dirs):
        params_file = osp.join(root_dir, log_dir, params_basename)
        try:
            with open(params_file) as f:
                params = json.load(f)
        except IOError:
            data.append({"log_dir": log_dir})
            continue
        params_keys = params_keys | set(params.keys())
        datum = params
        datum["log_dir"] = log_dir

        log_file = osp.join(root_dir, log_dir, log_basename)
        ext = osp.splitext(log_basename)[1]
        try:
            if ext == ".csv":
                log = pandas.read_csv(log_file)
            elif ext == ".json":
                with open(log_file) as f:
                    log = pandas.DataFrame(json.load(f))
            else:
                data.append(datum)
                continue
            if log.empty:
                raise pandas.errors.EmptyDataError
        except (IOError, pandas.errors.EmptyDataError):
            datum[key_epoch] = 0
            datum[key_iteration] = 0
            data.append(datum)
            continue

        if key_epoch not in log.columns:
            log[key_epoch] = 0

        datum[key_epoch] = log[key_epoch].max()
        datum[key_iteration] = log[key_iteration].max()
        datum[key_elapsed_time] = datetime.timedelta(
            seconds=int(round(log[key_elapsed_time].max()))
        )

        now = datetime.datetime.now()
        mtime = datetime.datetime.fromtimestamp(osp.getmtime(log_file))
        seconds = (now - mtime).total_seconds()
        datum["updated_at"] = datetime.timedelta(seconds=int(round(seconds)))

        for col in log.columns:
            if col in [key_epoch, key_iteration, key_elapsed_time]:
                continue

            key = "{} (max)".format(col)
            log_keys.add(key)
            index = log[col].idxmax()
            if math.isnan(index):
                break
            datum[key] = (
                log[col][index],
                (log[key_epoch][index], log[key_iteration][index]),
            )

            key = "{} (min)".format(col)
            log_keys.add(key)
            index = log[col].idxmin()
            datum[key] = (
                log[col][index],
                (log[key_epoch][index], log[key_iteration][index]),
            )

        data.append(datum)

    summary_keys = [key_epoch, key_iteration, key_elapsed_time, "updated_at"]
    summary_keys += sorted(params_keys)
    summary_keys += sorted(log_keys)

    df = pandas.DataFrame(data=data, columns=["log_dir"] + summary_keys)

    if sort is not None:
        key = sort
        ascending = True
        if key.startswith("-"):
            key = key[1:]
            ascending = False
        df = df.sort_values(by=key, ascending=ascending)

    # format
    for col in df.columns:
        if col.endswith(" (min)") or col.endswith(" (max)"):
            key = col[:-6]
            values = []
            for value in df[col].values:
                if isinstance(value, tuple):
                    values.append((float_format.format(value[0]), value[1],))
                else:
                    values.append((value, (0, 0)))
            df[col] = values
        elif df[col].dtype.name == "timedelta64[ns]":
            values = []
            for value in df[col]:
                try:
                    value = datetime.timedelta(seconds=value.total_seconds())
                    value = str(value)
                except ValueError:
                    pass
                values.append(value)
            df[col] = values
        else:
            df[col] = df[col].astype(str)

    if q is not None:
        try:
            df = df.query(q)
        except Exception:
            pass

    return df, summary_keys, params_keys, log_keys



# flake8: noqa

import argparse
import os.path as osp
import re
import sys
import textwrap
import warnings

import tabulate
import yaml

from . import __version__
from .parser import parse


def main():
    config_file = ".logtable"
    if osp.exists(config_file):
        print(" * Config file: {}".format(config_file))
        with open(".logtable") as f:
            default_args = yaml.safe_load(f)
    else:
        default_args = {}

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--version", "-V", action="store_true", help="show version and exit",
    )
    parser.add_argument(
        "--logdir",
        "-l",
        default=default_args.pop("logdir", "logs"),
        help="logs dir",
    )
    parser.add_argument(
        "--exclude",
        "-e",
        nargs="+",
        default=default_args.pop("exclude", []),
        help="exclude keys",
    )
    parser.add_argument(
        "--include",
        "-i",
        nargs="+",
        default=default_args.pop("include", []),
        help="include keys",
    )
    parser.add_argument("--keys", action="store_true")
    parser.add_argument(
        "--significant-figures",
        "-s",
        type=int,
        default=default_args.pop("significant_figures", 3),
        help="significant figures",
    )
    parser.add_argument(
        "--index",
        type=int,
        nargs="+",
        default=default_args.pop("index", None),
        help="index filtering",
    )
    parser.add_argument(
        "--column",
        "-c",
        type=int,
        default=default_args.pop("column", 30),
        help="maximum column width (0 for infinite)",
    )
    parser.add_argument(
        "--multi-column",
        action="store_true",
        default=default_args.pop("multi_column", False),
        help="prefer multi-column to vertically shorten the table",
    )
    parser.add_argument(
        "--params-basename",
        default=default_args.pop("params_basename", "params.json"),
        help="params or args of the experiment",
    )
    parser.add_argument(
        "--log-basename",
        default=default_args.pop("log_basename", "log.csv"),
        help="log of the experiment",
    )
    parser.add_argument(
        "--key-epoch",
        default=default_args.pop("key_epoch", "epoch"),
        help="key name of the epoch in the log file",
    )
    parser.add_argument(
        "--key-iteration",
        default=default_args.pop("key_iteration", "iteration"),
        help="key name of the iteration in the log file",
    )
    parser.add_argument(
        "--key-elapsed-time",
        default=default_args.pop("key_elapsed_time", "elapsed_time"),
        help="key name of the elapsed_time in the log file",
    )
    args = parser.parse_args()

    if args.version:
        print("logtable {}".format(__version__))
        sys.exit(0)

    if default_args:
        warnings.warn(
            "Config file .logtable has unsupported keys: {}".format(
                list(default_args.keys())
            )
        )

    float_format = "{:.%dg}" % args.significant_figures

    root_dir = osp.abspath(args.logdir)
    df, summary_keys, _, log_keys = parse(
        root_dir,
        float_format=float_format,
        params_basename=args.params_basename,
        log_basename=args.log_basename,
        key_epoch=args.key_epoch,
        key_iteration=args.key_iteration,
        key_elapsed_time=args.key_elapsed_time,
    )

    if args.index is not None:
        df = df.iloc[args.index]

    print(" * Log directory: {}".format(args.logdir))

    summary_keys = ["log_dir"] + summary_keys

    summary_keys_ = []
    for key in summary_keys:
        for excl in args.exclude:
            if re.match(excl, key):
                if any(re.match(incl, key) for incl in args.include):
                    summary_keys_.append(key)
                break
        else:
            summary_keys_.append(key)
    summary_keys = summary_keys_

    if args.keys:
        for key in summary_keys:
            print(key)
        sys.exit(0)

    headers = [""]
    for key in summary_keys:
        key_splits = key.replace("/", "/\n").replace("_", "_\n").splitlines()
        if args.column:
            key_splits = [
                x
                for key_split in key_splits
                for x in textwrap.wrap(key_split, width=args.column)
            ]
        key = "\n".join(key_splits)
        headers.append(key)
        if args.multi_column and key in log_keys:
            headers.append("")

    if df.empty:
        df.columns = summary_keys

    rows = []
    for index, df_row in df[summary_keys].iterrows():
        row = [index]
        for key, x in zip(summary_keys, df_row):
            if isinstance(x, tuple):
                assert len(x) == 2
                if args.multi_column:
                    row.extend(x)
                else:
                    x = "\n".join(str(xi) for xi in x)
                    row.append(x)
            else:
                x = str(x)
                x = textwrap.wrap(x, width=args.column) if args.column else [x]
                x = "\n".join(x)
                row.append(x)
        rows.append(row)

    table = tabulate.tabulate(
        tabular_data=rows,
        headers=headers,
        tablefmt="fancy_grid",
        stralign="center",
        showindex=False,
        disable_numparse=True,
    )
    print(table)
